const std = @import("std");
const SDL = @import("sdl2");
const sched = @import("scheduler.zig");
const emu = @import("chip8.zig");
const display = @import("display.zig");

const Timer = std.time.Timer;
const Chip8 = emu.Chip8;
const Display = display.Display;
const Scheduler = sched.Scheduler;

const OPS_PER_HZ = sched.OPS_PER_HZ;
const CHIP8_WIDTH = display.WIDTH;
const CHIP8_HEIGHT = display.HEIGHT;
const WINDOW_WIDTH = CHIP8_WIDTH * display.SCALE;
const WINDOW_HEIGHT = CHIP8_HEIGHT * display.SCALE;
const PIXELBUF_LEN = display.PIXELBUF_LEN;

pub fn main() anyerror!void {
    const heap = std.heap.page_allocator;
    var arg_it = std.process.args();

    // Skip process name
    _ = arg_it.skip();

    const os_string = try arg_it.next(heap) orelse {
        std.debug.warn("Expected first argument to be path to CHIP-8 rom\n", .{});
        return;
    };
    const rom_path = try std.fs.path.resolve(heap, &[_][]const u8{os_string});

    if (SDL.SDL_Init(SDL.SDL_INIT_VIDEO | SDL.SDL_INIT_EVENTS | SDL.SDL_INIT_AUDIO) < 0) {
        sdlPanic();
    }
    defer SDL.SDL_Quit();

    const window = SDL.SDL_CreateWindow(
        "Zig CHIP-8 Emulator",
        SDL.SDL_WINDOWPOS_CENTERED,
        SDL.SDL_WINDOWPOS_CENTERED,
        WINDOW_WIDTH,
        WINDOW_HEIGHT,
        SDL.SDL_WINDOW_SHOWN,
    ) orelse sdlPanic();
    defer _ = SDL.SDL_DestroyWindow(window);

    const renderer = SDL.SDL_CreateRenderer(window, -1, SDL.SDL_RENDERER_ACCELERATED) orelse sdlPanic();
    defer _ = SDL.SDL_DestroyRenderer(renderer);

    const texture = SDL.SDL_CreateTexture(renderer, SDL.SDL_PIXELFORMAT_RGBA8888, SDL.SDL_TEXTUREACCESS_STATIC, 64, 32);
    defer SDL.SDL_DestroyTexture(texture);

    var pixels: [PIXELBUF_LEN]u8 = [_]u8{0x00} ** PIXELBUF_LEN;

    var scheduler = Scheduler.new(heap, &pixels);
    var chip8 = try Chip8.fromFile(heap, &scheduler, rom_path);

    const frametime = 1_000_000_000 / OPS_PER_HZ;
    var timer = Timer.start() catch unreachable;

    emuloop: while (true) {
        var e: SDL.SDL_Event = undefined;
        while (SDL.SDL_PollEvent(&e) != 0) {
            if (e.type == SDL.SDL_QUIT) {
                break :emuloop;
            }
        }

        _ = SDL.SDL_UpdateTexture(texture, null, &pixels, CHIP8_WIDTH * @sizeOf(u32));
        _ = SDL.SDL_RenderCopy(renderer, texture, null, null);
        SDL.SDL_RenderPresent(renderer);

        sched.run_until(&scheduler, &chip8, scheduler.now() + OPS_PER_HZ);

        while (true) {
            const diff = timer.read();

            if (diff >= frametime) {
                timer.reset();
                break;
            }
        }
    }
}

fn sdlPanic() noreturn {
    const str = @as(?[*:0]const u8, SDL.SDL_GetError()) orelse "unknown error";
    @panic(std.mem.sliceTo(str, 0));
}
